@@ -1,10 +1,14 @@
 /******************************************************************************
- * hypercall-x86_32.h
+ * hypercall-x86_64.h
  * 
  * Copied from XenLinux.
  * 
  * Copyright (c) 2002-2004, K A Fraser
  * 
+ * 64-bit updates:
+ *   Benjamin Liu <benjamin.liu@intel.com>
+ *   Jun Nakajima <jun.nakajima@intel.com>
+ * 
  * This file may be distributed separately from the Linux kernel, or
  * incorporated into other software packages, subject to the following license:
  * 
@@ -27,18 +31,16 @@
  * IN THE SOFTWARE.
  */
 
-#ifndef __HYPERCALL_X86_32_H__
-#define __HYPERCALL_X86_32_H__
+#ifndef __HYPERCALL_X86_64_H__
+#define __HYPERCALL_X86_64_H__
 
 #include <xen/xen.h>
 #include <xen/sched.h>
-#include <xen/nmi.h>
 #include <mini-os/mm.h>
 
-typedef struct { unsigned long pte_low, pte_high; } pte_t;
+typedef struct { unsigned long pte; } pte_t;
 
-#define __pte(x) ({ unsigned long long _x = (x);        \
-    ((pte_t) {(unsigned long)(_x), (unsigned long)(_x>>32)}); })
+#define __pte(x) ((pte_t) { (x) } )
 
 #define __STR(x) #x
 #define STR(x) __STR(x)
@@ -61,7 +63,7 @@ extern char hypercall_page[PAGE_SIZE];
 	long __res, __ign1;					\
 	asm volatile (						\
 		"call hypercall_page + ("STR(__HYPERVISOR_##name)" * 32)"\
-		: "=a" (__res), "=b" (__ign1)			\
+		: "=a" (__res), "=D" (__ign1)			\
 		: "1" ((long)(a1))				\
 		: "memory" );					\
 	(type)__res;						\
@@ -72,7 +74,7 @@ extern char hypercall_page[PAGE_SIZE];
 	long __res, __ign1, __ign2;				\
 	asm volatile (						\
 		"call hypercall_page + ("STR(__HYPERVISOR_##name)" * 32)"\
-		: "=a" (__res), "=b" (__ign1), "=c" (__ign2)	\
+		: "=a" (__res), "=D" (__ign1), "=S" (__ign2)	\
 		: "1" ((long)(a1)), "2" ((long)(a2))		\
 		: "memory" );					\
 	(type)__res;						\
@@ -83,7 +85,7 @@ extern char hypercall_page[PAGE_SIZE];
 	long __res, __ign1, __ign2, __ign3;			\
 	asm volatile (						\
 		"call hypercall_page + ("STR(__HYPERVISOR_##name)" * 32)"\
-		: "=a" (__res), "=b" (__ign1), "=c" (__ign2), 	\
+		: "=a" (__res), "=D" (__ign1), "=S" (__ign2), 	\
 		"=d" (__ign3)					\
 		: "1" ((long)(a1)), "2" ((long)(a2)),		\
 		"3" ((long)(a3))				\
@@ -93,28 +95,30 @@ extern char hypercall_page[PAGE_SIZE];
 
 #define _hypercall4(type, name, a1, a2, a3, a4)			\
 ({								\
-	long __res, __ign1, __ign2, __ign3, __ign4;		\
+	long __res, __ign1, __ign2, __ign3;			\
 	asm volatile (						\
+		"movq %7,%%r10; "				\
 		"call hypercall_page + ("STR(__HYPERVISOR_##name)" * 32)"\
-		: "=a" (__res), "=b" (__ign1), "=c" (__ign2),	\
-		"=d" (__ign3), "=S" (__ign4)			\
+		: "=a" (__res), "=D" (__ign1), "=S" (__ign2),	\
+		"=d" (__ign3)					\
 		: "1" ((long)(a1)), "2" ((long)(a2)),		\
-		"3" ((long)(a3)), "4" ((long)(a4))		\
-		: "memory" );					\
+		"3" ((long)(a3)), "g" ((long)(a4))		\
+		: "memory", "r10" );				\
 	(type)__res;						\
 })
 
 #define _hypercall5(type, name, a1, a2, a3, a4, a5)		\
 ({								\
-	long __res, __ign1, __ign2, __ign3, __ign4, __ign5;	\
+	long __res, __ign1, __ign2, __ign3;			\
 	asm volatile (						\
+		"movq %7,%%r10; movq %8,%%r8; "			\
 		"call hypercall_page + ("STR(__HYPERVISOR_##name)" * 32)"\
-		: "=a" (__res), "=b" (__ign1), "=c" (__ign2),	\
-		"=d" (__ign3), "=S" (__ign4), "=D" (__ign5)	\
+		: "=a" (__res), "=D" (__ign1), "=S" (__ign2),	\
+		"=d" (__ign3)					\
 		: "1" ((long)(a1)), "2" ((long)(a2)),		\
-		"3" ((long)(a3)), "4" ((long)(a4)),		\
-		"5" ((long)(a5))				\
-		: "memory" );					\
+		"3" ((long)(a3)), "g" ((long)(a4)),		\
+		"g" ((long)(a5))				\
+		: "memory", "r10", "r8" );			\
 	(type)__res;						\
 })
 
@@ -155,12 +159,11 @@ HYPERVISOR_stack_switch(
 
 static inline int
 HYPERVISOR_set_callbacks(
-	unsigned long event_selector, unsigned long event_address,
-	unsigned long failsafe_selector, unsigned long failsafe_address)
+	unsigned long event_address, unsigned long failsafe_address, 
+	unsigned long syscall_address)
 {
-	return _hypercall4(int, set_callbacks,
-			   event_selector, event_address,
-			   failsafe_selector, failsafe_address);
+	return _hypercall3(int, set_callbacks,
+			   event_address, failsafe_address, syscall_address);
 }
 
 static inline int
@@ -189,9 +192,7 @@ static inline long
 HYPERVISOR_set_timer_op(
 	uint64_t timeout)
 {
-	unsigned long timeout_hi = (unsigned long)(timeout>>32);
-	unsigned long timeout_lo = (unsigned long)timeout;
-	return _hypercall2(long, set_timer_op, timeout_lo, timeout_hi);
+	return _hypercall1(long, set_timer_op, timeout);
 }
 
 static inline int
@@ -210,9 +211,9 @@ HYPERVISOR_get_debugreg(
 
 static inline int
 HYPERVISOR_update_descriptor(
-	uint64_t ma, uint64_t desc)
+	unsigned long ma, unsigned long word)
 {
-	return _hypercall4(int, update_descriptor, ma, ma>>32, desc, desc>>32);
+	return _hypercall2(int, update_descriptor, ma, word);
 }
 
 static inline int
@@ -233,15 +234,14 @@ static inline int
 HYPERVISOR_update_va_mapping(
 	unsigned long va, pte_t new_val, unsigned long flags)
 {
-	return _hypercall4(int, update_va_mapping, va,
-			   new_val.pte_low, new_val.pte_high, flags);
+	return _hypercall3(int, update_va_mapping, va, new_val.pte, flags);
 }
 
 static inline int
 HYPERVISOR_event_channel_op(
-	int cmd, void *op)
+       int cmd, void *op)
 {
-	return _hypercall2(int, event_channel_op, cmd, op);
+    return _hypercall2(int, event_channel_op, cmd, op);
 }
 
 static inline int
@@ -276,8 +276,8 @@ static inline int
 HYPERVISOR_update_va_mapping_otherdomain(
 	unsigned long va, pte_t new_val, unsigned long flags, domid_t domid)
 {
-	return _hypercall5(int, update_va_mapping_otherdomain, va,
-			   new_val.pte_low, new_val.pte_high, flags, domid);
+	return _hypercall4(int, update_va_mapping_otherdomain, va,
+			   new_val.pte, flags, domid);
 }
 
 static inline int
@@ -294,12 +294,19 @@ HYPERVISOR_vcpu_op(
 	return _hypercall3(int, vcpu_op, cmd, vcpuid, extra_args);
 }
 
+static inline int
+HYPERVISOR_set_segment_base(
+	int reg, unsigned long value)
+{
+	return _hypercall2(int, set_segment_base, reg, value);
+}
+
 static inline int
 HYPERVISOR_suspend(
 	unsigned long srec)
 {
-	return _hypercall3(int, sched_op, SCHEDOP_shutdown,
-			   SHUTDOWN_suspend, srec);
+	struct sched_shutdown shutdown = { .reason = SHUTDOWN_suspend };
+	return _hypercall3(int, sched_op, SCHEDOP_shutdown, &shutdown, srec);
 }
 
 static inline int
@@ -324,7 +331,13 @@ HYPERVISOR_domctl(
 	return _hypercall1(int, domctl, op);
 }
 
-#endif /* __HYPERCALL_X86_32_H__ */
+static inline unsigned long
+HYPERVISOR_hvm_op(int op, void *arg)
+{
+	return _hypercall2(unsigned long, hvm_op, op, arg);
+}
+
+#endif /* __HYPERCALL_X86_64_H__ */
 
 /*
  * Local variables:
