@@ -10,21 +10,20 @@
  * we have to worry about.
  */
 
-#include <xen/config.h>
 #include <xen/delay.h>
 #include <xen/time.h>
 #include <asm/msr.h>
 #include <asm/processor.h>
 
-void __udelay(unsigned long usecs)
+void udelay(unsigned long usecs)
 {
     unsigned long ticks = usecs * (cpu_khz / 1000);
     unsigned long s, e;
 
-    s = rdtsc();
+    s = rdtsc_ordered();
     do
     {
-        rep_nop();
-        e = rdtsc();
+        cpu_relax();
+        e = rdtsc_ordered();
     } while ((e-s) < ticks);
 }
