@@ -1,9 +1,7 @@
 #ifndef __XEN_EFI_H__
 #define __XEN_EFI_H__
 
-#ifndef __ASSEMBLY__
 #include <xen/types.h>
-#endif
 
 #define EFI_INVALID_TABLE_ADDR (~0UL)
 
@@ -23,15 +21,23 @@ struct efi {
 
 extern struct efi efi;
 
-#ifndef __ASSEMBLY__
-
 union xenpf_efi_info;
 union compat_pf_efi_info;
 
 struct xenpf_efi_runtime_call;
 struct compat_pf_efi_runtime_call;
 
+#if defined(CONFIG_X86) || defined(CONFIG_ARM)
 bool efi_enabled(unsigned int feature);
+#else
+static inline bool efi_enabled(unsigned int feature)
+{
+    return false;
+}
+#endif
+
+extern bool efi_secure_boot;
+
 void efi_init_memory(void);
 bool efi_boot_mem_unused(unsigned long *start, unsigned long *end);
 bool efi_rs_using_pgtables(void);
@@ -45,6 +51,4 @@ int efi_runtime_call(struct xenpf_efi_runtime_call *op);
 int efi_compat_get_info(uint32_t idx, union compat_pf_efi_info *info);
 int efi_compat_runtime_call(struct compat_pf_efi_runtime_call *op);
 
-#endif /* !__ASSEMBLY__ */
-
 #endif /* __XEN_EFI_H__ */
